/*
 * ArduinoExampleMacros:
 *
 * Arduino sketch that accepts commands from a user interface on
 * a Reach Technology SLCD43 Development Kit display and then
 * drive an analog pulse with a settable pulse profile (PWM duty
 * cycle and duration).
 *
 * This sketch is an example of a high-level control application
 * that receives information from macros running on the display. 
 * The information is pulse profile parameters, trigger events
 * and screen change notifications.  The sketch does not manage
 * the buttons or sliders on the display.  The sketch does need
 * to know which screen (main or alt) the display is using so it
 * can interpret the pulse profile information correctly.
 *
 * Timer2 on the Arduio is used to drive a PWM signal to create
 * the output pulse.  The PWM output from timer2 on OC2B (pin 9
 * on Mega2560) is filtered with an RC network to convert the
 * output into an analog signal.  The duty cycle is varied to
 * adjust the pulse amplitude.  The pulse duty cycle and duration
 * are based on information received from the display.
 *
 * The low-level pulse profile settings are displayed on the
 * screen each time the pulse is triggered (either by receiving
 * a trigger event from the display or by the footswitch input).
 *
 * An LED with 330 ohm current limiting resistor connected to
 * pin 12 will turn on for the selected pulse time each time a
 * pulse is triggered.
 */

// compiler-provided definitions
// __AVR_ATmega328P__ - Uno
// __AVR_ATmega1280__ - Mega
// __AVR_ATmega2560__ - Mega 2560


// notifications from display, see SLCDx Software Reference
#define  USER_MSG          '@'  // indicates start of user-defined message
#define  PULSE_PROFILE     'p'
#define  TRIGGER           't'
#define  SCREEN_CHANGE     's'

// maximum number of commands sent without getting a reply
#define  MAX_CMD_OUTSTANDING  2

// duty cycle definitions assuming 'fast PWM' config
#define  DUTY_CYCLE_MAX     255
#define  DUTY_CYCLE_20PCNT  ((DUTY_CYCLE_MAX*20)/100)
//#define  DUTY_CYCLE_40PCNT  ((DUTY_CYCLE_MAX*40)/100)
//#define  DUTY_CYCLE_60PCNT  ((DUTY_CYCLE_MAX*60)/100)
//#define  DUTY_CYCLE_99PCNT  ((DUTY_CYCLE_MAX*99)/100)
#define  MIN_DUTY_CYCLE     DUTY_CYCLE_20PCNT

// duration literals
#define  MIN_DURATION      50      // ms
#define  MAX_DURATION      200     // ms

// length of serial input buffer
#define  BUFF_SIZE        100

// pin definitions
#define  FOOTSWITCH_IN    2
#define  PWM_OUT          9  // timer2 OC2B output
#define  DEBUG_LED        12

// serial IO port mapping literals
#define  DISPLAY_SIO          Serial1
#define  DISPLAY_SERIALEVENT  serialEvent1
#define  DEBUG_SIO            Serial    // to use IDE's Serial Monitor

// debug support
#undef  SHOW_DEBUG
#ifdef  SHOW_DEBUG
// these are undefined if SHOW_DEBUG is undefined,
// otherwise may be defined or not as needed
#define  USER_MESSAGE_DEBUG
#define  PULSE_PROFILE_DEBUG
#define  DEBUG_SCREEN_NAME
#undef   DEBUG_CMD_OVERLAP
#endif

// misc
#define  FOOTSWITCH_DEBOUNCE  50  // time in milliseconds

// variables and flags
boolean  main_screen;
byte footswitch_state;
byte last_footswitch_state;
boolean trigger_set;
boolean trigger_holdoff;
byte pulse_duty_cycle;
int pulse_duration;
char buffer[BUFF_SIZE];
int  buffer_index;
char  replies_req;
boolean message_ready;
boolean serial_error;

// function prototypes
void handle_pulse_profile(void);
int decode_ASCII(char buffer[], int *index, char delim);
void print_pulse_profile(void);
void do_pulse(void);
void handle_footswitch(void);
void wait_for_display(void);
void DISPLAY_SERIALEVENT(void);


/*
 * setup:
 *
 * the required Arduino sketch setup() function.
 */
 void setup()
{

  // timer2 output pin
  pinMode(PWM_OUT, OUTPUT);

  // footswitch input, with internal 20k pull-up
  pinMode(FOOTSWITCH_IN, INPUT);
  digitalWrite(FOOTSWITCH_IN, HIGH);

  // debug LED
  pinMode(DEBUG_LED, OUTPUT);

  // init the serial port and send a few <return>s
  // to get the display's attention
  DISPLAY_SIO.begin(115200);
  DISPLAY_SIO.print('\r');
  DISPLAY_SIO.print('\r');
  DISPLAY_SIO.print('\r');

#ifdef  SHOW_DEBUG
  DEBUG_SIO.begin(9600);
  DEBUG_SIO.println("Showing debug");
#endif

  footswitch_state = digitalRead(FOOTSWITCH_IN);
  last_footswitch_state = footswitch_state;
  trigger_set = false;
  trigger_holdoff = false;
  pulse_duty_cycle = MIN_DUTY_CYCLE;
  pulse_duration = MIN_DURATION;
  buffer_index = 0;
  replies_req = 0;
  message_ready = false;
  serial_error = false;

  // tell the screen to initialize itself
  DISPLAY_SIO.println("m power_on");
  replies_req++;
  wait_for_display();

}


/*
 * loop:
 *
 * the required Arduino loop() function.
 */
void loop()
{
  int  i, num_chars, index, pp_dc, pp_dur;
  
  // check for display notifications
  if (message_ready)
  {
//    DEBUG_SIO.println("Handling a message");
    message_ready = false;
    if (buffer[0] == USER_MSG)
    {
#ifdef  USER_MESSAGE_DEBUG
      DEBUG_SIO.println(buffer);
#endif
      switch (buffer[1])
      {
        case PULSE_PROFILE:
          handle_pulse_profile();
          break;
        case TRIGGER:
          trigger_set = true;
          break;
        case SCREEN_CHANGE:
          // screen change messages are in the form "@s:name\r"
          // where 'name' is "main" or "alt".  we'll take the
          // short-cut of only looking at the first character 
          // of the screen name.  then, if needed, code can do
          // something like:
          //    if (main_screen)
          //    {
          //    }
          //    else
          //    {
          //    }

          if (buffer[3] == 'm')
          {
            main_screen = true;
          }
          else if (buffer[3] == 'a')
          {
            main_screen = false;
          }
          else
          {
#ifdef  DEBUG_SCREEN_NAME
            DISPLAY_SIO.println("z");
            DISPLAY_SIO.print("t \"Bad @s command: ");
            DISPLAY_SIO.print(buffer);
            DISPLAY_SIO.println("\" 100 100");
#endif
          }
          break;
        default:
          // ignore unknown user messages
          break;
      }
    }
  }

  // check for footswitch trigger
  handle_footswitch();

  // trigger pulse if flag is set
  if (trigger_set)
  {
    print_pulse_profile();
    trigger_set = false;
    do_pulse();
  }

}


/*
 * handle_pulse_profile:
 *
 * called to handle pulse profile user commands.  decodes the two
 * profile values depending on which screen (main or alt) is active.
 */
void handle_pulse_profile(void)
{
  int  index, pp_dc, pp_dur;
  
  // pulse profile messages are in the form "@p:dc,dur\r"
  // buffer[2] is ':'
  index = 3;
  pp_dc = decode_ASCII(buffer, &index, ',');
  index++;    // skip delimiter
  pp_dur = decode_ASCII(buffer, &index, ',');
  // pp_dc is the duty cycle percentage (1-99)
  // and pp_dur is the duration in ms (50-200)
  pulse_duty_cycle = map(pp_dc,
                          0, 100,
                          MIN_DUTY_CYCLE, DUTY_CYCLE_MAX-1);
  if (pp_dur < MIN_DURATION)
  {
    pulse_duration = MIN_DURATION;
  }
  else if (pp_dur > MAX_DURATION)
  {
    pulse_duration = MAX_DURATION;
  }
  else pulse_duration = pp_dur;  // OK to use directly

}


/*
 * decode_ASCII:
 *
 * called to decod the ASCII numeric value at buffer[index].
 * returns the decoded value, and leaves index set to the
 * delimiter or the zero terminator.
 *
 * assumes that there are no more than three ASCII characters.
 */
int decode_ASCII(char buffer[], int *index, char delim)
{
  int  val, l_index;

  l_index = *index;
  val = buffer[l_index++] - '0';
  if ((buffer[l_index] != delim) && (buffer[l_index] != 0))
  {
    val = (val * 10) + (buffer[l_index++] - '0');
  }
  if ((buffer[l_index] != delim) && (buffer[l_index] != 0))
  {
    val = (val * 10) + (buffer[l_index++] - '0');
  }

#ifdef  PULSE_PROFILE_DEBUG
  DEBUG_SIO.print("decode_ASCII: ");
  DEBUG_SIO.print(val, DEC);
  DEBUG_SIO.print(" ending index: ");
  DEBUG_SIO.println(l_index, DEC);
#endif

  *index = l_index;
  return(val);

}


/*
 * print_pulse_profile:
 *
 * called to print the current low-level pulse profile
 */
void print_pulse_profile(void)
{

  wait_for_display();
  DISPLAY_SIO.print("m display_pulse_profile ");
  DISPLAY_SIO.print(pulse_duty_cycle, DEC);
  DISPLAY_SIO.print(" ");
  DISPLAY_SIO.println(pulse_duration, DEC);
  replies_req++;

}


/*
 * do_pulse:
 *
 * called to issue a pulse as specified by the current
 * pulse profile settings.
 *
 * NOTE: The direct use of processor registers makes this code
 * specific to a processor function (output OSC2B) instead of
 * an Arduino function (e.g. analogWrite on pin 3) and this
 * processor function may appear on different Arduino pins
 * depending on which processor is used.  The ifdefs at the
 * beginning of this file allow this same code to be used on a
 * Uno and on a Mega 2560.
 */
void do_pulse(void)
{

  digitalWrite(DEBUG_LED, 1);

  // set timer2 for fast PWM, start timer
  TCCR2A = _BV(COM2A1) | _BV(COM2B1) | _BV(WGM21) | _BV(WGM20);
  TCCR2B = _BV(CS20);  // prescler = 1

  // the pulse
  OCR2B = pulse_duty_cycle;
  delay(pulse_duration);

  OCR2B = 1;          // minimum duty cycle
  TCCR2A = _BV(0);    // stop timer

  digitalWrite(DEBUG_LED, 0);

}


/*
 * handle_footswitch:
 *
 * called to monitor and debounce the footswitch input.
 * sets trigger_set when a valid footswitch action is
 * detected (falling edge).  waits for rising edge to
 * reenable footswitch action.
 */
void handle_footswitch(void)
{

  footswitch_state = digitalRead(FOOTSWITCH_IN);
  if (footswitch_state != last_footswitch_state)
  {
    if ((footswitch_state == LOW)  && (!trigger_holdoff))
    {
      delay(FOOTSWITCH_DEBOUNCE);
      footswitch_state = digitalRead(FOOTSWITCH_IN);
      // if footswitch is still low, OK to trigger pulse
      if (footswitch_state == LOW)
      {
        trigger_set = true;
        trigger_holdoff = true;
      }
    }
    last_footswitch_state = footswitch_state;
  }
  else
  {
    // fotswitch has been released, allow new trigger
    if (footswitch_state == HIGH)
    {
      trigger_holdoff = false;
    }
  }

}


/*
 * wait_for_display:
 *
 * called to wait for display to reply to previous commands.
 * used in lieu of XON/XOFF flow control.  application is able
 * to send up to MAX_CMD_OUTSTANDING overlapping commands before
 * needing to wait for display to reply.
 */
void wait_for_display(void)
{

  while (replies_req >= MAX_CMD_OUTSTANDING)
  {
#ifdef  DEBUG_CMD_OVERLAP
    DEBUG_SIO.print("Stalled: ");
    DEBUG_SIO.println(replies_req, DEC);
#endif
    // wait until display catches up with commands
    DISPLAY_SERIALEVENT();  // must call manually
    delay(10);
  }

}


/*
 * DISPLAY_SERIALEVENT:
 *
 * called when serial data is available.  reads serial data into
 * the buffer and sets the line_ready flag when a <return> char
 * has been received.
 *
 * NOTE: This function is called by the Arduino environment each
 * time loop() is executed, after loop() is called.  If loop()
 * is waiting for serial data it will need to call this function
 * manually.
 */
void DISPLAY_SERIALEVENT(void)
{
  char  serial_char;

  while (DISPLAY_SIO.available())
  {
    serial_char = DISPLAY_SIO.read();
  
#undef  DEBUG_SERIAL
#ifdef  DEBUG_SERIAL
    // send same data to debug serial port
    DEBUG_SIO.print(serial_char);
    if (serial_char == '\r')
    {
      DEBUG_SIO.print("<return>");
      DEBUG_SIO.print("\n");
    }
#endif

    if (serial_char == '>')
    {
      if (replies_req)
      {
        replies_req--;
      }
    }
    // lines starting with:
    //   '!' indicates command failure
    //   '^' indicates display buffer full
    //   '?' indicates transmission line error
    //   '#' indicates CRC error
    //   ':' indicates human-readable information
    else if ((buffer_index == 0) &&
              ((serial_char == '!') ||
               (serial_char == '^') ||
               (serial_char == '?') ||
               (serial_char == '#') ||
               (serial_char == ':')))
    {
      serial_error = true;
      // additional handling depends on user application
    }
    else if (serial_char == '\r')
    {
      // end of line
      buffer[buffer_index] = '\0';  // string terminator
      if (buffer_index > 0)
      {
#ifdef  DEBUG_SERIAL
        DEBUG_SIO.println("Got a message");
#endif
        message_ready = true;
        buffer_index = 0;
      }
    }
    else
    {
      // store character
      buffer[buffer_index] = serial_char;
      buffer_index++;
    }
    
  }  /* while */

}


